import sys

if sys.platform != "cli":
    raise(RuntimeError, "This module is only intended to be run in Rhino Python")

import clr

clr.AddReference('Grasshopper, Culture=neutral, PublicKeyToken=dda4f5ec2cd80803')

import System
import Rhino
import Grasshopper as gh
import sys
import re

from System.Collections import IEnumerable, IEnumerator


class namespace_object(object):
    pass


def __make_function_uo__(helper):
    def component_function(*args, **kwargs):
        comp = helper.proxy.CreateInstance()
        comp.ClearData()
        if args:
            for i, arg in enumerate(args):
                if arg is None: continue
                param = comp.Params.Input[i]
                param.PersistentData.Clear()
                if hasattr(arg, '__iter__'):  # TODO deal with polyline, str
                    [param.AddPersistentData(a) for a in arg]
                else:
                    param.AddPersistentData(arg)
        if kwargs:
            for param in comp.Params.Input:
                name = param.Name.lower()
                if name in kwargs:
                    param.PersistentData.Clear()
                    arg = kwargs[name]
                    if hasattr(arg, '__iter__'):  # TODO deal with polyline, str
                        [param.AddPersistentData(a) for a in arg]
                    else:
                        param.AddPersistentData(arg)
        doc = gh.Kernel.GH_Document()
        doc.AddObject(comp, False, 0)
        comp.CollectData()
        comp.ComputeData()
        output = helper.create_output(comp.Params)
        comp.ClearData()
        doc.Dispose()
        return output

    return component_function


class function_helper(object):
    def __init__(self, proxy, name):
        self.proxy = proxy
        self.return_type = None

    def create_output(self, params, output_values=None):
        if not output_values:
            output_values = []
            for output in params.Output:
                data = output.VolatileData.AllData(True)
                # We could call Value, but ScriptVariable seems to do a better job
                v = [x.ScriptVariable() for x in data]
                if len(v) < 1:
                    output_values.append(None)
                elif len(v) == 1:
                    output_values.append(v[0])
                else:
                    output_values.append(v)
        if len(output_values) == 1: return output_values[0]
        if self.return_type is None:
            names = [output.Name.lower() for output in params.Output]
            try:
                self.return_type = namedtuple('Output', names, rename=True)
            except:
                self.return_type = False
        if not self.return_type: return output_values
        return self.return_type(*output_values)

    def runfast(self, args, kwargs):
        return False, None


def __build_module_uo():
    core_module = sys.modules[__name__]
    translate_from = u"|+-*\u2070\u00B9\u00B2\u00B3\u2074\u2075\u2076\u2077\u2078\u2079"
    translate_to = "X__x0123456789"
    transl = dict(zip(translate_from, translate_to))

    def regex_helper(match):
        if match.group() in transl:
            return transl[match.group()]
        return ''

    def function_description(description, params):
        rc = ['', description, "Input:"]
        for param in params.Input:
            s = "\t{0} [{1}] - {2}"
            if param.Optional:
                s = "\t{0} (in, optional) [{1}] - {2}"
            rc.append(s.format(param.Name.lower(), param.TypeName, param.Description))
        if params.Output.Count == 1:
            param = params.Output[0]
            rc.append("Returns: [{0}] - {1}".format(param.TypeName, param.Description))
        elif params.Output.Count > 1:
            rc.append("Returns:")
            for out in params.Output:
                s = "\t{0} [{1}] - {2}"
                rc.append(s.format(out.Name.lower(), out.TypeName, out.Description))
        return '\n'.join(rc)

    for obj in gh.Instances.ComponentServer.ObjectProxies:
        if obj.Exposure == gh.Kernel.GH_Exposure.hidden or obj.Obsolete:
            continue

        t = clr.GetClrType(gh.Kernel.IGH_Component)
        library_id = obj.LibraryGuid
        assembly = gh.Instances.ComponentServer.FindAssembly(library_id)
        name = obj.Desc.Name

        if "LEGACY" in name or "#" in name:
            continue
        name = re.sub("[^a-zA-Z0-9]", regex_helper, name)
        if not name[0].isalpha():
            name = 'x' + name

        m = core_module
        if assembly is None:
            # UserObjects
            pass
        elif not (t.IsAssignableFrom(obj.Type)):
            # Specialized inputs/outputs and misconfigured components
            # Discard
            continue
        elif not assembly.IsCoreLibrary:
            # Compiled components, leave them to ghpythonlib
            continue

        function = __make_function_uo__(function_helper(obj, name))
        try:
            setattr(m, name, function)
            a = m.__dict__[name]
            a.__name__ = name
            comp = obj.CreateInstance()
            a.__doc__ = function_description(obj.Desc.Description, comp.Params)
        except Exception as err:
            Rhino.RhinoApp.WriteLine(str(err))
            Rhino.Runtime.HostUtils.ExceptionReport("ghpythonlib.components.py|" + name,
                                                    err.clsException)


__build_module_uo()



import json
import logging
import os
import subprocess

logger = logging.getLogger("ghpythonremote.helpers")


def get_python_path(location=None):
    if location is None or location == "":
        return get_python_from_windows_path()

    if os.path.exists(location):
        logger.debug(" Directly trying python executable at {!s}\n".format(location))
        return get_python_from_path(location)

    try:
        [method, env_name] = location.split("://")
    except ValueError as e:
        logger.warning(
            "Location format for {!s} did not match expected format: "
            '"method://env_name"({!s}).\n'.format(location, e.message)
            + "Falling back to getting python path from windows %PATH%.\n"
        )
        return get_python_from_windows_path()
    if method == "conda":
        return get_python_from_conda_env(env_name)

    logger.warning(
        "Method {!s} for location not implemented.\n".format(method)
        + "Falling back to getting python path from windows %PATH%.\n"
    )
    return get_python_from_windows_path()


def get_python_from_path(path):
    path = os.path.normpath(os.path.realpath(path))
    if os.path.isdir(path):
        path = os.path.join(path, "python.exe")
    if os.path.isfile(path) and os.access(path, os.X_OK):
        return path
    else:
        logger.warning("Path {!s} is not executable.\n".format(path))
        return get_python_from_windows_path()


def get_python_from_windows_path():
    try:
        python_exe = subprocess.check_output(["where", "python"]).split("\n")[0].strip()
        return python_exe
    except (OSError, subprocess.CalledProcessError) as e:
        logger.error(
            "Unable to find a python installation in your windows %PATH%. "
            + "Are you running Windows with python accessible in your path?\n"
        )
        raise e


def get_python_from_conda_env(env_name):
    try:
        envs = json.loads(subprocess.check_output(["conda", "env", "list", "--json"]))[
            "envs"
        ]
    except OSError:
        logger.warning(
            "conda not found in your windows %PATH%, cannot fetch environment by "
            + "name.\nFalling back to getting python path from windows %PATH%.\n"
        )
        return get_python_path()
    except subprocess.CalledProcessError:
        logger.warning(
            "conda env list failed, cannot fetch environment by name.\n"
            + "Falling back to getting python path from windows %PATH%.\n"
        )
        return get_python_path()

    env_dir = [path for path in envs if os.path.split(path)[-1] == env_name]
    if len(env_dir) > 1:
        logger.warning(
            "Found several environments with target name; selecting first one."
        )

    try:
        python_exe = os.path.join(env_dir[0], "python.exe")
        return python_exe
    except IndexError:
        logger.warning(
            "Environment {!s} was not found in your conda list of "
            "environments.\n".format(env_name)
            + "Falling back to getting python path from windows %PATH%.\n"
        )
        return get_python_from_windows_path()


def get_extended_env_path_conda(python_exe):
    # Conda stores useful DLLs in these paths, add them to the path like they would be
    # by conda activate
    new_env = os.environ.copy()
    conda_folders = [
        "",
        "/Library/mingw-w64/bin",
        "/Library/usr/bin",
        "/Library/bin",
        "/Scripts",
        "/bin",
    ]
    add_path = [
        os.path.normpath(os.path.dirname(python_exe) + folder)
        for folder in conda_folders
    ]
    new_env["PATH"] = os.pathsep.join(add_path) + os.pathsep + new_env["PATH"]
    return new_env


def get_rhino_ironpython_path(location=None):
    if location is None or location == "":
        return get_ironpython_from_appdata()

    if type(location) == int:
        logger.debug(
            " Looking for IronPython installation of Rhino version {!s}.\n".format(
                location
            )
        )
        return get_ironpython_from_appdata(location)

    if os.path.isdir(location):
        logger.debug(" Directly using IronPython lib folder at {!s}\n".format(location))
        return get_ironpython_from_path(location)

    logger.warning(
        " Path {!s} is not a directory or does not exist.\n".format(location)
        + " " * 9
        + "Falling back to getting IronPython lib folder path from windows %APPDATA%.\n"
    )
    return get_ironpython_from_appdata()


DEFAULT_RHINO_VERSION = 6


def get_ironpython_from_appdata(rhino_version=DEFAULT_RHINO_VERSION):
    appdata_path = os.getenv("APPDATA", "")
    if rhino_version == 6:
        ironpython_settings_path = os.path.join(
            appdata_path,
            "McNeel",
            "Rhinoceros",
            "6.0",
            "Plug-ins",
            "IronPython (814d908a-e25c-493d-97e9-ee3861957f49)",
            "settings",
        )
    elif rhino_version == 5:
        ironpython_settings_path = os.path.join(
            appdata_path,
            "McNeel",
            "Rhinoceros",
            "5.0",
            "Plug-ins",
            "IronPython (814d908a-e25c-493d-97e9-ee3861957f49)",
            "settings",
        )
    else:
        logger.warning(
            ' Unknown Rhino version "{!s}". Defaulting to Rhino {!s}.\n'.format(
                rhino_version, DEFAULT_RHINO_VERSION
            )
        )
        return get_ironpython_from_appdata()

    if rhino_version == 5:
        ghpython_version_path = os.path.join(
            ironpython_settings_path, "ghpy_version.txt"
        )

        if os.path.isfile(ghpython_version_path):
            with open(ghpython_version_path) as ghpython_version:
                ghpython_version_list = ghpython_version.readline().split(".")
            try:
                ghpython_version_tuple = tuple(int(x) for x in ghpython_version_list)
                if ghpython_version_tuple < (0, 6, 0, 3):
                    logger.warning(
                        " ghpy_version.txt indicates obsolete version {!s}.\n".format(
                            ".".join(ghpython_version_list)
                        )
                        + " " * 9
                        + "Please install version 0.6.0.3 or superior from "
                        + "http://www.food4rhino.com/app/ghpython\n"
                    )
                logger.info(
                    " Found ghpython version {!s} in {!s}\n".format(
                        ".".join(ghpython_version_list), ironpython_settings_path
                    )
                )
            except ValueError:
                logger.warning(
                    " Could not parse ghpy_version.txt file installation found in "
                    "{!s}.\n".format(ironpython_settings_path)
                    + " " * 9
                    + "Was ghpython installed and opened in Grasshopper at least once "
                    + "on this machine?\n"
                )
        else:
            logger.warning(
                " No ghpy_version.txt file installation found in {!s}.\n".format(
                    ironpython_settings_path
                )
                + " " * 9
                + "Was ghpython installed and opened in Grasshopper at least once on "
                + "this machine?\n"
            )
    if rhino_version == 6:
        # Rhino 6 does not have a ghpy_version.txt file, check that there is a
        # ghpythonlib __init__.py
        ghpythonlib_init_path = os.path.join(
            ironpython_settings_path, "lib", "ghpythonlib", "__init__.py"
        )

        if not os.path.isfile(ghpythonlib_init_path):
            logger.warning(
                " No ghpythonlib package found in {!s}.\n".format(
                    os.path.join(ironpython_settings_path, "lib")
                )
                + " " * 9
                + "Was Grasshopper for Rhino 6 opened at least once on this machine?\n"
            )

    ironpython_lib_path = os.path.join(ironpython_settings_path, "lib")
    if not os.path.isdir(ironpython_lib_path):
        logger.error(
            " IronPython lib directory for Rhinoceros not found in {!s}.\n".format(
                ironpython_settings_path
            )
            + " " * 7
            + "Please provide a full path to one of the folders in your IronPython for "
            + "Rhinoceros path.\n"
            + " " * 7
            + "These folders are listed in the settings of the window opened with the "
            + "command `_EditPythonScript` in Rhinoceros.\n"
        )
        raise RuntimeError(
            "No IronPython lib folder found in %APPDATA%\\McNeel\\Rhinoceros"
        )
    logger.info(" Found IronPython lib folder {!s}\n".format(ironpython_lib_path))

    return ironpython_lib_path


def get_ironpython_from_path(location):
    return location



import inspect
import logging
from os import path

import ghpythonremote
from ghpythonremote.connectors import PythonToGrasshopperRemote

logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.DEBUG)

# TODO: This could be made a console script in setup.py, with some additional config
if __name__ == "__main__":
    ROOT = path.abspath(path.dirname(inspect.getfile(ghpythonremote)))
    rhino_file_path = path.join(ROOT, "examples", "curves.3dm")
    rpyc_server_py = path.join(ROOT, "ghcompservice.py")

    with PythonToGrasshopperRemote(
        rhino_file_path, rpyc_server_py, rhino_ver=6, timeout=60
    ) as py2gh:
        # Stuff that we can reach
        rghcomp = py2gh.gh_remote_components  # Named Grasshopper compiled components
        rghuo = py2gh.gh_remote_userobjects  # Named Grasshopper user objects
        rgh = py2gh.connection  # Represents the remote instance root object
        Rhino = (
            rgh.modules.Rhino
        )  # rgh.modules.something is like ``import something`` on the remote
        rs = rgh.modules.rhinoscriptsyntax  # Same

        # Opening a Rhino file
        readopt = Rhino.FileIO.FileReadOptions()
        readopt.BatchMode = True
        Rhino.RhinoDoc.ReadFile(
            rhino_file_path, readopt
        )  # Or pass in a first argument to py2gh to open a file

        # Doing stuff in that file
        type_curve = Rhino.DocObjects.ObjectType.Curve
        curves = Rhino.RhinoDoc.ActiveDoc.Objects.FindByObjectType(type_curve)
        curves_id = tuple(
            c.Id for c in curves
        )  # rhinoscriptsyntax doesn't like mutable objects through the connection
        gh_curves = rs.coerceguidlist(curves_id)
        # Call a GH component
        print(sum(rghcomp.Area(gh_curves)[0]))
        # Call a GH user object, previously created with the name "TestClusterGHPR"
        print(rghuo.TestClusterGHPR(3, y=4))



import errno
import logging
import os
import platform
import socket
import subprocess
from time import sleep

try:
    import _winreg as winreg
except ImportError:
    import winreg

from ghpythonremote import rpyc
from .helpers import get_python_path, get_extended_env_path_conda

logger = logging.getLogger("ghpythonremote.connectors")


class GrasshopperToPythonRemote:
    def __init__(
        self,
        rpyc_server_py,
        python_exe=None,
        location=None,
        timeout=60,
        max_retry=3,
        port=None,
        log_level=logging.WARNING,
        working_dir=None,
    ):
        if python_exe is None:
            self.python_exe = get_python_path(location)
        else:
            if location is not None:
                logger.debug(
                    "python_exe and env_name specified at the same time, ignoring "
                    "env_name."
                )
            self.python_exe = python_exe
        self.env = get_extended_env_path_conda(self.python_exe)
        self.rpyc_server_py = rpyc_server_py
        self.timeout = timeout
        self.retry = 0
        self.max_retry = max(0, max_retry)
        self.log_level = log_level
        self.working_dir = working_dir
        if port is None:
            self.port = _get_free_tcp_port()
        else:
            self.port = port
        self.python_popen = self._launch_python()
        self.connection = self._get_connection()
        self.py_remote_modules = self.connection.root.getmodule

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Cleanup the connection on error and exit.

        Parameters
        ----------
        exc_type : Error
            Type of the exception that caused the __exit__.
        exc_val : str
            Value of the exception that caused the __exit__.
        exc_tb : type
            Exception log.

        Returns
        -------
        True if the connection was successfully closed."""
        try:
            if exc_type:
                logger.error("{!s}, {!s}, {!s}".format(exc_type, exc_val, exc_tb))
        except NameError:
            pass
        self.close()
        return True

    def run_py_function(self, module_name, function_name, *nargs, **kwargs):
        """Run a specific Python function on the remote, with Python crash handling."""
        remote_module = self.py_remote_modules(module_name)
        function = getattr(remote_module, function_name)
        function_output = kwargs.pop("function_output", None)

        try:
            result = function(*nargs, **kwargs)
        except (socket.error, EOFError):
            self._rebuild_py_remote()
            return self.run_py_function(*nargs, **kwargs)

        if function_output is not None:
            try:
                result = result[function_output]
            except NameError:
                pass
        return result

    def close(self):
        if not self.connection.closed:
            logger.info("Closing connection.")
            self.connection.close()
        if self.python_popen.poll() is None:
            logger.info("Closing Python.")
            self.python_popen.terminate()

    def _launch_python(self):
        logger.debug("Using python executable: {!s}".format(self.python_exe))
        logger.debug("Using rpyc_server module: {!s}".format(self.rpyc_server_py))
        logger.debug("Using port: {}".format(self.port))
        logger.debug("Using log_level: {!s}".format(self.log_level))
        logger.debug("Using working_dir: {!s}".format(self.working_dir))
        assert self.python_exe is not "" and self.python_exe is not None
        assert self.rpyc_server_py is not "" and self.rpyc_server_py is not None
        assert self.port is not "" and self.port is not None
        assert self.log_level is not "" and self.log_level is not None
        python_call = '"{!s}" "{!s}" "{}" "{!s}"'.format(
            self.python_exe, self.rpyc_server_py, self.port, self.log_level
        )
        cwd = self.working_dir
        python_popen = subprocess.Popen(
            python_call,
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            cwd=cwd,
            env=self.env,
        )
        return python_popen

    def _get_connection(self):
        connection = None
        logger.info("Connecting...")
        for i in range(self.timeout):
            try:
                if not connection:
                    logger.debug(
                        "Connecting. Timeout in {:d} seconds.".format(self.timeout - i)
                    )
                    connection = rpyc.utils.factory.connect(
                        "localhost",
                        self.port,
                        service=rpyc.core.service.ClassicService,
                        config={"sync_request_timeout": None},
                        ipv6=False,
                        keepalive=True,
                    )
                else:
                    logger.debug(
                        "Found connection, testing. Timeout in {:d} seconds.".format(
                            self.timeout - i
                        )
                    )
                    connection.ping(timeout=1)
                    logger.debug("Connection ok, returning.")
                    logger.info("Connected.")
                    return connection
            except socket.error as e:
                if self.python_popen.poll() is not None:
                    raise RuntimeError(
                        "Remote python {!s} failed on launch. ".format(self.python_exe)
                        + "Does the remote python have rpyc installed?"
                    )
                if i == self.timeout - 1 or not e.errno == errno.ECONNREFUSED:
                    raise RuntimeError(
                        "Could not connect to remote python {!s}. ".format(
                            self.python_exe
                        )
                        + "Does the remote python have rpyc installed?"
                    )
                sleep(1)
            except (
                rpyc.core.protocol.PingError,
                rpyc.core.async_.AsyncResultTimeout,
            ) as e:
                logger.debug(str(e))
                raise e

    def _rebuild_py_remote(self):
        if self.retry < self.max_retry:
            logger.info("Lost Rhino connection, retrying.")
            self.retry += 1
            self.close()
            [self.rhino_popen, self.connection, self.gh_remote] = [None, None, None]
            logger.info("Waiting 10 seconds.")
            sleep(10)
            self.python_popen = self._launch_python()
            self.connection = self._get_connection()
        else:
            raise RuntimeError(
                "Lost connection to Python, and reconnection attempts limit ({:d}) "
                "reached. Exiting.".format(self.max_retry)
            )


class PythonToGrasshopperRemote:
    """Creates a remote Rhino/IronPython instance (with Grasshopper functions)
    connected to a local python engine.
    
    The local instance will be able to import all objects from the Rhino IronPython
    engine, as well as Grasshopper components. Rhino will appear frozen on a python
    script it is reading.
        
    Parameters
    ----------
    rhino_file_path : str
        Absolute file path to a Rhino .3dm file to open in the remote Rhino. Can be
        empty.
    rpyc_server_py : str
        Absolute path to the ghcompservice.py module that launches the server on the
        remote.
    rhino_ver : int
        A Rhino version to use, from 5 to 7. Overridden by rhino_exe. Defaults to 6.
    rhino_exe : str
        Absolute path to the Rhino executable. By default, fetches from the windows
        registry the Rhino install with the same bitness as the platform, and version
        given by rhino_ver.
    timeout : int
        Number of seconds to wait for Rhino and IronPython to startup.
    max_retry : int
        Number of times Rhino will be restarted if it crashes, before declaring the
        connection dead.
    
    Examples
    --------
    >>> ROOT = os.path.abspath(os.path.join(os.path.curdir, '..'))
    >>> rhino_file_path = os.path.join(ROOT, 'examples', 'curves.3dm')
    >>> rpyc_server_py = os.path.join(ROOT, 'ghcompservice.py')
    >>> with PythonToGrasshopperRemote(
    >>>     rhino_file_path, rpyc_server_py, rhino_ver=6, timeout=60
    >>> ) as py2gh:
    >>>     rghcomp = py2gh.gh_remote_components
    >>>     rgh = py2gh.connection
    >>>     Rhino = rgh.modules.Rhino
    >>>     rs = rgh.modules.rhinoscriptsyntax
    >>>     # Do stuff with all this
    >>>     # See CPython_to_GH.py for a longer example
    """

    def __init__(
        self,
        rhino_file_path,
        rpyc_server_py,
        rhino_ver=6,
        rhino_exe=None,
        timeout=60,
        max_retry=3,
        port=None,
    ):
        if rhino_exe is None:
            self.rhino_exe = self._get_rhino_path(version=rhino_ver)
        else:
            self.rhino_exe = rhino_exe
        self.rhino_file_path = rhino_file_path
        self.rpyc_server_py = rpyc_server_py
        self.timeout = timeout
        self.retry = 0
        self.max_retry = max(0, max_retry)
        if port is None:
            self.port = _get_free_tcp_port()
        else:
            self.port = port
        self.rhino_popen = self._launch_rhino()
        self.connection = self._get_connection()
        self.gh_remote_components = self.connection.root.ghcomp
        self.gh_remote_userobjects = self.connection.root.ghuo

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Cleanup the connection on error and exit.
        
        Parameters
        ----------
        exc_type : Error
            Type of the exception that caused the __exit__.
        exc_val : str
            Value of the exception that caused the __exit__.
        exc_tb : type
            Exception log.
            
        Returns
        -------
        True if the connection was successfully closed."""
        try:
            if exc_type:
                logger.error("{!s}, {!s}, {!s}".format(exc_type, exc_val, exc_tb))
        except NameError:
            pass
        self.close()
        return True

    def run_gh_component(self, component_name, *nargs, **kwargs):
        """Run a specific Grasshopper component on the remote, with Rhino crash
        handling.
        """
        is_cluster = kwargs.pop("is_cluster", False)
        # TODO: improve ghcomp to get clusters the same way we get compiled components,
        # thus removing the need for a custom getter
        component = self.gh_remote_components(component_name, is_cluster=is_cluster)
        component_output = kwargs.pop("component_output", None)

        try:
            result = component(*nargs, **kwargs)
        except (socket.error, EOFError):
            self._rebuild_gh_remote()
            return self.run_gh_component(*nargs, **kwargs)

        if component_output is not None:
            try:
                result = result[component_output]
            except NameError:
                pass
        return result

    def close(self):
        if not self.connection.closed:
            logger.info("Closing connection.")
            self.connection.close()
        if self.rhino_popen.poll() is None:
            logger.info("Closing Rhino.")
            self.rhino_popen.terminate()

    @staticmethod
    def _get_rhino_path(version=6, preferred_bitness="same"):
        rhino_reg_key_path = None
        version_str = "{!s}.0".format(version)
        if platform.architecture()[0] == "64bit":
            if preferred_bitness == "same" or preferred_bitness == "64":
                if version == 5:
                    version_str += "x64"
                rhino_reg_key_path = r"SOFTWARE\McNeel\Rhinoceros\{}\Install".format(
                    version_str
                )
                if version < 5:
                    rhino_reg_key_path = None
            elif preferred_bitness == "32":
                rhino_reg_key_path = (
                    r"SOFTWARE\WOW6432Node\McNeel\Rhinoceros\{}\Install"
                )
                rhino_reg_key_path = rhino_reg_key_path.format(version_str)
        elif platform.architecture()[0] == "32bit":
            if preferred_bitness == "same" or preferred_bitness == "32":
                rhino_reg_key_path = r"SOFTWARE\McNeel\Rhinoceros\{}\Install".format(
                    version_str
                )
            if version > 5:
                rhino_reg_key_path = None

        if rhino_reg_key_path is None:
            logger.error(
                "Did not understand Rhino version ({!s}) and bitness ({!s}) options "
                "for platform {!s}.".format(
                    version, preferred_bitness, platform.machine()
                )
            )

        # In Python 3, OpenKey might throw a FileNotFoundError, which is not defined in
        # Python 2. Just pretend to work around that
        try:
            FileNotFoundError
        except NameError:
            FileNotFoundError = IOError
        try:
            rhino_reg_key = winreg.OpenKey(
                winreg.HKEY_LOCAL_MACHINE, rhino_reg_key_path
            )
            rhino_path = winreg.QueryValueEx(rhino_reg_key, "Path")[0]
        except (FileNotFoundError, OSError) as e:
            logger.error(
                "Unable to find Rhino installation in registry. Are you running "
                "Windows with Rhinoceros installed?"
            )
            raise e
        return os.path.join(rhino_path, "Rhino.exe")

    def _launch_rhino(self):
        assert self.rhino_exe is not "" and self.rhino_exe is not None
        assert self.rpyc_server_py is not "" and self.rpyc_server_py is not None
        assert self.port is not "" and self.port is not None
        rhino_call = [
            '"' + self.rhino_exe + '"',
            "/nosplash",
            "/notemplate",
            '/runscript="-_RunPythonScript ""{!s}"" {!s} -_Exit "'.format(
                self.rpyc_server_py, self.port
            ),
        ]
        if self.rhino_file_path:
            rhino_call.append(self.rhino_file_path)
        # Default escaping in subprocess.line2cmd does not work here,
        # manually convert to string
        rhino_call = " ".join(rhino_call)
        rhino_popen = subprocess.Popen(
            rhino_call, stdout=subprocess.PIPE, stdin=subprocess.PIPE
        )
        return rhino_popen

    def _get_connection(self):
        connection = None
        logger.info("Connecting...")
        for i in range(self.timeout):
            try:
                if not connection:
                    logger.debug(
                        "Connecting. Timeout in {:d} seconds.".format(self.timeout - i)
                    )
                    connection = rpyc.utils.factory.connect(
                        "localhost",
                        self.port,
                        service=rpyc.core.service.ClassicService,
                        config={"sync_request_timeout": None},
                        ipv6=False,
                        keepalive=True,
                    )
                else:
                    logger.debug(
                        "Found connection, testing. Timeout in {:d} seconds.".format(
                            self.timeout - i
                        )
                    )
                    connection.ping(timeout=1)
                    logger.debug("Connection ok, returning.")
                    logger.info("Connected.")
                    return connection
            except (
                socket.error,
                rpyc.core.protocol.PingError,
                rpyc.core.async_.AsyncResultTimeout,
            ) as e:
                if e is socket.error and not e.errno == errno.ECONNREFUSED:
                    raise
                if i == self.timeout - 1:
                    raise
                elif e is socket.error:
                    sleep(1)

    def _rebuild_gh_remote(self):
        if self.retry < self.max_retry:
            logger.info("Lost Rhino connection, retrying.")
            self.retry += 1
            self.close()
            [self.rhino_popen, self.connection, self.gh_remote] = [None, None, None]
            logger.info("Waiting 10 seconds.")
            sleep(10)
            self.rhino_popen = self._launch_rhino()
            self.connection = self._get_connection()
            self.gh_remote_components = self.connection.root.get_component
        else:
            raise RuntimeError(
                "Lost connection to Rhino, and reconnection attempts limit ({:d}) "
                "reached. Exiting.".format(self.max_retry)
            )


def _get_free_tcp_port():
    tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    tcp.bind(("", 0))
    addr, port = tcp.getsockname()
    tcp.close()
    return port



import logging
import os
from shutil import copy, rmtree
import subprocess
import sys

from .helpers import get_rhino_ironpython_path

logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.INFO)

if __name__ == "__main__":
    location = None
    if len(sys.argv) > 1:
        location = sys.argv[1]
        try:
            location = int(location)
        except ValueError:
            pass

    rhino_ironpython_path = get_rhino_ironpython_path(location=location)

    # Install the package to Rhino IronPython Users lib
    package_name = "gh-python-remote"
    pip_cmd = [
        sys.executable,
        "-m",
        "pip",
        "install",
        "--upgrade",
        "--target=" + rhino_ironpython_path,
        "--no-compile",
        "--ignore-requires-python",
        package_name,
    ]
    print(
        "The gh-python-remote package will be installed in Rhino IronPython with the "
        "command:"
    )
    print(" ".join(pip_cmd))
    subprocess.check_call(pip_cmd)

    # Get the Grasshopper libraries base dir
    gh_userobjects_path = os.path.join(
        os.getenv("APPDATA", ""), "Grasshopper", "UserObjects"
    )
    if os.path.isdir(gh_userobjects_path):
        dest_dir_path = os.path.join(gh_userobjects_path, package_name)
    else:
        logging.warning(
            'Could not find Grasshopper "UserObjects" special folder; '
            + "example files will be installed to default temp dir."
        )
        dest_dir_path = os.path.join(os.getenv("TEMP", "C:\\"), package_name)
    # Get the filepath in the installed package
    source_dir_path = os.path.dirname(os.path.realpath(__file__))
    # Send the files there
    if os.path.isdir(dest_dir_path):
        rmtree(dest_dir_path)
    os.mkdir(dest_dir_path)
    copy_pairs = [
        (
            "examples",
            [
                "examples/curves.3dm",
                "examples/GH_python_remote.ghx",
                "examples/GH_python_remote_plt_example.ghx",
                "examples/CPython_to_GH.py",
            ],
        ),
        ("", ["ghcluster/GHPythonRemote.ghuser"]),
    ]
    for copy_pair in copy_pairs:
        dest = os.path.join(dest_dir_path, copy_pair[0])
        if not os.path.isdir(dest):
            os.mkdir(dest)
        for file_rel_path in copy_pair[1]:
            source = os.path.join(source_dir_path, *file_rel_path.split("/"))
            copy(source, dest)
    logging.info("Copied example files to {!s}".format(dest_dir_path))



import logging
import sys

from ghpythonremote import rpyc
from rpyc.utils.server import OneShotServer


class PythonService(rpyc.ClassicService):
    def on_connect(self, conn):
        logger.info("Incoming connection.")
        super(PythonService, self).on_connect(conn)

    def on_disconnect(self, conn):
        logger.info("Disconnected.")


if __name__ == "__main__":

    if len(sys.argv) >= 3:
        log_level = sys.argv[2]
        try:
            log_level = int(log_level)
        except (TypeError, ValueError):
            log_level = getattr(logging, log_level, logging.WARNING)
    else:
        log_level = logging.WARNING
    if len(sys.argv) >= 2:
        port = sys.argv[1]
    else:
        port = 18871
    try:
        port = int(port)
    except (TypeError, ValueError):
        port = 18871

    # Log everything that happens on the Python server in the console
    logger = logging.getLogger()
    logger.setLevel(log_level)
    ch = logging.StreamHandler()
    ch.setLevel(log_level)
    formatter = logging.Formatter(
        "\n%(asctime)s - %(name)s - %(levelname)s -\n%(message)s"
    )
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    logger = logging.getLogger("ghpythonremote.pythonservice")
    logger.info("Starting server...")
    server = OneShotServer(
        PythonService,
        hostname="localhost",
        port=port,
        listener_timeout=None,
        logger=logger,
    )
    server.start()



from ghpythonremote import rpyc
from rpyc.utils.server import OneShotServer


class GhcompService(rpyc.ClassicService):
    def on_connect(self, conn):
        print("Incoming connection.")
        super(GhcompService, self).on_connect(conn)
        import ghpythonlib.components as ghcomp
        from ghpythonremote import ghuserobjects as ghuo

        self.ghcomp = ghcomp
        self.ghuo = ghuo

    def on_disconnect(self, conn):
        print("Disconnected.")


if __name__ == "__main__":
    import rhinoscriptsyntax as rs

    port = rs.GetInteger("Server bind port", 18871, 1023, 65535)

    server = OneShotServer(
        GhcompService, hostname="localhost", port=port, listener_timeout=None
    )
    server.start()



import inspect
import sys

import rpyc

if sys.platform == "cli":
    # Some compatibility fixes for IronPython
    rpyc.core.brine.IMM_INTS = dict((i, bytes([i + 0x50])) for i in range(-0x30, 0xA0))

    def BYTES_LITERAL(text):
        return bytes(text)

    rpyc.lib.compat.BYTES_LITERAL = BYTES_LITERAL

    # TODO: This will not be found in Rhino 5!
    # TODO: This will not work if in IronPython outside Rhino!
    # TODO: This loads the gh components early?? not needed if not accessing them??
    # TODO: Remove that when ghpythonlib.components.__namedtuple.__getattr__ is fixed
    import ghpythonlib.components

    def get_id_pack(obj):
        """introspects the given "local" object, returns id_pack as expected by BaseNetref

        The given object is "local" in the sense that it is from the local cache. Any object in the local cache exists
        in the current address space or is a netref. A netref in the local cache could be from a chained-connection.
        To handle type related behavior properly, the attribute `__class__` is a descriptor for netrefs.

        So, check thy assumptions regarding the given object when creating `id_pack`.
        """

        # TODO: Remove that when ghpythonlib.components.__namedtuple.__getattr__ is fixed
        if hasattr(obj, "____id_pack__") and not isinstance(
            obj, ghpythonlib.components.__namedtuple
        ):
            # netrefs are handled first since __class__ is a descriptor
            return obj.____id_pack__
        # str(obj).split(':')[0] == "Microsoft.Scripting.Actions.NamespaceTracker" should also work
        elif (
            inspect.ismodule(obj)
            or getattr(obj, "__name__", None) == "module"
            or str(type(obj)) == "<type 'namespace#'>"
        ):
            # TODO: not sure about this, need to enumerate cases in units
            if isinstance(obj, type):  # module
                obj_cls = type(obj)
                name_pack = "{0}.{1}".format(obj_cls.__module__, obj_cls.__name__)
                return (name_pack, id(type(obj)), id(obj))
            else:
                if inspect.ismodule(obj) and obj.__name__ != "module":
                    if obj.__name__ in sys.modules:
                        name_pack = obj.__name__
                    else:
                        name_pack = "{0}.{1}".format(
                            obj.__class__.__module__, obj.__name__
                        )
                elif inspect.ismodule(obj):
                    name_pack = "{0}.{1}".format(obj.__module__, obj.__name__)
                    print(name_pack)
                elif hasattr(obj, "__module__"):
                    name_pack = "{0}.{1}".format(obj.__module__, obj.__name__)
                else:
                    obj_cls = type(obj)
                    name_pack = "{0}".format(obj.__name__)
                return (name_pack, id(type(obj)), id(obj))
        elif not inspect.isclass(obj):
            name_pack = "{0}.{1}".format(
                obj.__class__.__module__, obj.__class__.__name__
            )
            return (name_pack, id(type(obj)), id(obj))
        else:
            name_pack = "{0}.{1}".format(obj.__module__, obj.__name__)
            return (name_pack, id(obj), 0)

    rpyc.lib.get_id_pack = get_id_pack
    rpyc.core.netref.get_id_pack = get_id_pack
    rpyc.core.protocol.get_id_pack = get_id_pack

    # TODO: Remove that when ghpythonlib.components.__namedtuple.__getattr__ is fixed
    def _handle_inspect(self, id_pack):  # request handler
        obj = self._local_objects[id_pack]
        if hasattr(obj, "____conn__") and not isinstance(
            obj, ghpythonlib.components.__namedtuple
        ):
            # When RPyC is chained (RPyC over RPyC), id_pack is cached in local objects as a netref
            # since __mro__ is not a safe attribute the request is forwarded using the proxy connection
            # see issue #346 or tests.test_rpyc_over_rpyc.Test_rpyc_over_rpyc
            conn = self._local_objects[id_pack].____conn__
            return conn.sync_request(rpyc.core.consts.HANDLE_INSPECT, id_pack)
        else:
            return tuple(
                rpyc.lib.get_methods(
                    rpyc.core.netref.LOCAL_ATTRS, self._local_objects[id_pack]
                )
            )

    # TODO: Remove that when ghpythonlib.components.__namedtuple.__getattr__ is fixed
    rpyc.core.protocol.Connection._handle_inspect = _handle_inspect

    if sys.version_info < (2, 7, 5):

        def dump(obj):
            stream = []
            rpyc.core.brine._dump(obj, stream)
            return b"".join(map(bytes, stream))

        rpyc.core.brine.dump = dump
else:
    # This is only needed if the local is CPython and the remote is IronPython, doesn't
    # really hurt otherwise
    _netref_factory_orig = rpyc.core.protocol.Connection._netref_factory

    def _netref_factory_str(self, id_pack):
        return _netref_factory_orig(self, (str(id_pack[0]), id_pack[1], id_pack[2]))

    rpyc.core.protocol.Connection._netref_factory = _netref_factory_str



# To use a consistent encoding
from codecs import open
from os import path
from setuptools import setup, find_packages
import sys

if sys.version_info[0] > 2:
    sys.exit("Incompatible with Python 3. IronPython 2.7 from Rhino can only be "
             "connected to a Python 2 instance.")

here = path.abspath(path.dirname(__file__))

# Default version
__version__ = '0.1.0'
# Get __version__, __version_info__
execfile(path.join(here, 'ghpythonremote', 'version.py'))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()
# Append the CHANGELOG to it
with open(path.join(here, 'CHANGELOG.rst'), encoding='utf-8') as f:
    long_description += '\n\n====\n\n' + f.read()

setup(
    name='gh-python-remote',
    description=('GH Python Remote is a package to get Rhinoceros3D/Grasshopper and '
                 + 'Python to collaborate better: connect an external python instance'
                 + ' to Grasshopper, and vice-versa.'),
    long_description=long_description,
    long_description_content_type='text/x-rst',
    version=__version__,

    author='Pierre Cuvilliers',
    author_email='pcuvil@mit.edu',
    url='https://github.com/Digital-Structures/ghpythonremote',

    license='MIT',

    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Topic :: Software Development :: Interpreters',
        'Topic :: Scientific/Engineering',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 2 :: Only',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: IronPython',
    ],
    keywords='CAD design engineering scientific numpy scipy IronPython Rhinoceros3D',

    python_requires='>=2.6,<3',
    platforms=["Windows", ],
    zip_safe=False,

    use_2to3=False,

    packages=find_packages(),
    package_data={
        'ghpythonremote': [
            'examples/curves.3dm',
            'examples/GH_python_remote.ghx',
            'examples/GH_python_remote_plt_example.ghx',
            'examples/CPython_to_GH.py',
            'ghcluster/GHPythonRemote.ghuser',
        ],
    },
    data_files=[
    ],
    install_requires=[
        'rpyc==4.1.4'
    ],
    extras_require={
        'dev': ['check-manifest'],
        'test': ['coverage'],
    },
)
